//+------------------------------------------------------------------+
//|                                                       Unroll.mq5 |
//|                                                Maxim A.Kuznetsov |
//|                                          https://www.luxtrade.tk |
//+------------------------------------------------------------------+
#property copyright "Maxim A.Kuznetsov"
#property link      "https://www.luxtrade.tk"
#property version   "1.00"

#property description "Закрывает позы по кроссам, открывает эквивалентный объём через USD"
#property description "в результате : фиксируется PL по кроссам, снижается нагрузка на депо"
#property description "               валютная корзина в порядке (без встречных)"
#property description "на парах с usd могут образоваться локи - их надо потом закрыть/отрулить"
#property description "для новых позиций на usd выставляется TP в безубыток"

input bool ALLOW_TRADE=false;

#property script_show_inputs

#include <MT4Orders.mqh>
#ifdef __MQL5__
bool RefreshRates() {
   return false;
}
#endif

/// Информация по про валюту
class CurrencyInfo {
public:
   CurrencyInfo(string _name) {
      name=_name;
      volume=-1;
      profit=0; // Profit/Loss 
      loss=0;  // spread loss
   }
   ~CurrencyInfo(){
   }
public:
   double profit;     // 
   double loss;
   double volume;   
   string name;
};
CurrencyInfo *Currencys[];  // коллекция валют, для учёта объёмов

// получить символ из коллекции
CurrencyInfo *GetCurrency(string name)
{
   int index=-1;
   for(int pos=ArraySize(Currencys)-1;pos>=0;pos--) {
      if (Currencys[pos]==NULL) {
         index=pos;
         continue;
      }
      if (Currencys[pos].name==name) return Currencys[pos];
   }
   if (index==-1) {
      index=ArraySize(Currencys);
      if (ArrayResize(Currencys,index+1)!=index+1) return NULL;
   }
   Currencys[index]=new CurrencyInfo(name);
   return Currencys[index];
}
string GetSymbol(string one,string two)
{
   for(int pos=SymbolsTotal(false)-1;pos>=0;pos--) {
      string name=SymbolName(pos,false);
      if (name=="") continue;
      string base=SymbolInfoString(name,SYMBOL_CURRENCY_BASE);
      string quote=SymbolInfoString(name,SYMBOL_CURRENCY_PROFIT);
      if (one==base && two==quote) return name;
      if (one==quote && two==base) return name;
   }
   return "";
}
bool UnRoll(bool allowTrade=false)
{
   // закрывает позиции по кроссам, считаем итоговый объём по usd
   double profitOrders=0.0;   // посчитаем закрываемый PL
   int countOrders=0;      // и кол-во
   for(int pos=OrdersTotal()-1;pos>=0;pos--) {
      // выбрали ордер
      if (!OrderSelect(pos,SELECT_BY_POS,MODE_TRADES)) {
         Sleep(50);
         RefreshRates();
         if (!OrderSelect(pos,SELECT_BY_POS,MODE_TRADES)) {
            PrintFormat("OrderSelect() failed %d",GetLastError());
            return false;
         }
      }
      // только действующие рыночные
      if (OrderCloseTime()!=0) continue;
      int type=OrderType();
      if (type!=OP_BUY && type!=OP_SELL) continue;
      // только кроссы
      string symbol=OrderSymbol();
      string baseName=SymbolInfoString(symbol,SYMBOL_CURRENCY_BASE);
      string quoteName=SymbolInfoString(symbol,SYMBOL_CURRENCY_PROFIT);
      if (baseName==quoteName || baseName=="" || quoteName=="") continue;
      if (baseName=="USD" || quoteName=="USD") continue;
      CurrencyInfo *base=GetCurrency(baseName);
      CurrencyInfo *quote=GetCurrency(quoteName);
      if ( GetSymbol(base.name,"USD")==NULL || GetSymbol(quote.name,"USD")==NULL) {
         continue;
      }
      // закрываем
      PrintFormat("Close symbol=%s lots=%.3f direction=%s",symbol,OrderLots(),type==OP_BUY?"BUY":"SELL");
      if (!allowTrade) {
         profitOrders+=OrderProfit()+OrderSwap()+OrderCommission();
         countOrders++;
      }
      if (allowTrade) {
         if (!OrderClose(OrderTicket(),OrderLots(),OrderClosePrice(),5)) {
            Sleep(50);
            if (!OrderClose(OrderTicket(),OrderLots(),OrderClosePrice(),5)) {
               PrintFormat("OrderCLose() failed %d",GetLastError());
               continue;
            }
         }
         profitOrders+=OrderProfit()+OrderSwap()+OrderCommission();
         countOrders++;
      }
      // учитываем объём
      double contract=SymbolInfoDouble(symbol,SYMBOL_TRADE_CONTRACT_SIZE);
      if (contract<2) contract=10000;
      if (type==OP_BUY) {
         quote.volume+=OrderLots()*contract;
         base.volume+=OrderLots()*contract/OrderOpenPrice();    // OrderClosePrice()
      } else {
         quote.volume-=OrderLots()*contract;
         base.volume-=OrderLots()*contract/OrderOpenPrice();    // OrderClosePrice();
      }
      double orderProfit=OrderProfit()+OrderSwap()+OrderCommission();
      double orderLoss=SymbolInfoInteger(symbol,SYMBOL_SPREAD)*SymbolInfoDouble(symbol,SYMBOL_TRADE_TICK_VALUE)*OrderLots();
      base.profit+=orderProfit;
      base.loss+=orderLoss;
      quote.profit+=orderProfit;
      quote.loss+=orderLoss;
   }
   PrintFormat("Total %d orders with profit %.2f",countOrders,profitOrders);
   // напечатаем объёмы по валютам
   for(int pos=ArraySize(Currencys)-1;pos>=0;pos--) {
      CurrencyInfo *cur=Currencys[pos];
      if (cur==NULL) continue;         // перестарховка
      if (cur.volume==0) continue;
      PrintFormat("Currency=%s volume=%f direction=%s",cur.name,cur.volume,cur.volume>0?"BUY":"SELL");
   }
   // Открываем остатки на USD
   for(int pos=ArraySize(Currencys)-1;pos>=0;pos--) {
      CurrencyInfo *cur=Currencys[pos];
      if (cur==NULL) continue;         // перестарховка
      if (cur.volume==0) continue;
      // символ по которому открываться
      string symbol=GetSymbol(cur.name,"USD");  
      if (symbol=="") continue;
      double ask=SymbolInfoDouble(symbol,SYMBOL_ASK);
      double bid=SymbolInfoDouble(symbol,SYMBOL_BID);
      double contract=SymbolInfoDouble(symbol,SYMBOL_TRADE_CONTRACT_SIZE);
      // направление и объём в контрактах
      int type=OP_BUY;
      double lots=0;
      if (SymbolInfoString(symbol,SYMBOL_CURRENCY_PROFIT)=="USD") {
         // прямая котировка
         if (cur.volume>0) {
            type=OP_BUY;
            lots=cur.volume/ask/contract;
         } else {
            type=OP_SELL;
            lots=-cur.volume/bid/contract;
         }
      } else {
         // обратная котировка
         if (cur.volume>0) {
            type=OP_SELL;
            lots=cur.volume/contract;
         } else {
            type=OP_BUY;
            lots=-cur.volume/contract;
         }
      }
      PrintFormat("Open symbol=%s lots=%.2f direction=%s",symbol,lots,type==OP_BUY?"BUY":"SELL");
      if (lots<SymbolInfoDouble(symbol,SYMBOL_VOLUME_MIN)*0.8) {
         PrintFormat("Too small lots");
         continue;
      }
      if (allowTrade) {
         lots=NormalizeLots(lots,symbol);
         long ticket=OrderSend(symbol,type,lots,type==OP_BUY?ask:bid,5,0,0,"unroll",0);
         if (ticket<0) {
            PrintFormat("OrderSend() error %d",GetLastError());
         } else {
            if (!OrderSelect(ticket,SELECT_BY_TICKET)) {
               Sleep(50);
               if (!OrderSelect(ticket,SELECT_BY_TICKET)) {
                  PrintFormat("Unable to select new order!!");
                  return true;
               }
            }
            // выставить TP чтобы покрыть расходы
            int spread=(int)SymbolInfoInteger(symbol,SYMBOL_SPREAD);
            double tickValue=SymbolInfoDouble(symbol,SYMBOL_TRADE_TICK_VALUE);
            double pl=cur.profit-cur.loss-spread*tickValue*lots;
            int distance=(int)MathRound(MathAbs(pl)/(tickValue*lots));
            if (pl>0) distance=distance/2;
            int stopLevel=(int)SymbolInfoInteger(symbol,SYMBOL_TRADE_STOPS_LEVEL);
            if (stopLevel<2*spread) stopLevel=2*spread;
            if (distance<=stopLevel) distance=stopLevel+1;
            double tp=0;
            if (type==OP_BUY) {
               tp=NormalizeDouble(OrderOpenPrice()+distance*SymbolInfoDouble(symbol,SYMBOL_POINT),(int)SymbolInfoInteger(symbol,SYMBOL_DIGITS));
            } else {
               tp=NormalizeDouble(OrderOpenPrice()-distance*SymbolInfoDouble(symbol,SYMBOL_POINT),(int)SymbolInfoInteger(symbol,SYMBOL_DIGITS));
            }
            if (!OrderModify(ticket,OrderOpenPrice(),0,tp,0)) {
               Sleep(50);
               if (!OrderModify(ticket,OrderOpenPrice(),0,tp,0)) {
                  PrintFormat("Unable place TP %f for %d",tp,ticket);
               }
            }
         }
      }
   }
   return true;
}
double NormalizeLots(double lots,string symbol)
{
   double minLot=SymbolInfoDouble(symbol,SYMBOL_VOLUME_MIN);
   double maxLot=SymbolInfoDouble(symbol,SYMBOL_VOLUME_MAX);
   double lotStep=SymbolInfoDouble(symbol,SYMBOL_VOLUME_STEP);
   if (lots<=minLot) return minLot;
   if (lots>=maxLot) return maxLot;
   lots=minLot+MathRound((lots-minLot)/lotStep)*lotStep;
   return lots;
}
void OnStart()
{
   UnRoll(ALLOW_TRADE);
   for(int pos=ArraySize(Currencys)-1;pos>=0;pos--) {
      if (Currencys[pos]!=NULL) delete Currencys[pos];
   }
}
